"""
This file contains the utilities for all the I/O necessary in gulpy.

"""
import os
import numpy as np
from numba import njit

from oasislmf.pytools.getmodel.manager import get_damage_bins
from oasislmf.pytools.getmodel.common import oasis_float, areaperil_int
from oasislmf.pytools.gul.common import (
    ProbMean, damagecdfrec, damagecdfrec_stream,
    gulSampleslevelHeader, gulSampleslevelRec, oasis_float_to_int32_size,
    gulSampleslevelRec_size, gulSampleslevelHeader_size
)


def read_getmodel_stream(run_dir, stream_in, buff_size=65536):
    """Read the getmodel output stream yielding data event by event. 

    Args:
        run_dir (str): path to the run directory.
        stream_in (buffer-like): input stream, e.g. `sys.stdin.buffer`.
        buff_size (int): size in bytes of the read buffer. Default is 65536.

    Raises:
        ValueError: If the stream type is not 1.

    Yields:
        int32,  numpy.array[damagecdf], numpy.array[int], numpy.array[ProbMean]:
          event_id, array of damagecdf entries (areaperil_id, vulnerability_id) for this event,
          number of bins in all cdfs of event_id, all the cdfs used in event_id.
    """
    # determine stream type
    stream_type = np.frombuffer(stream_in.read(4), dtype='i4')

    # TODO: make sure the bit1 and bit 2-4 compliance is checked
    # see https://github.com/OasisLMF/ktools/blob/master/docs/md/CoreComponents.md
    if stream_type[0] != 1:
        raise ValueError(f"FATAL: Invalid stream type: expect 1, got {stream_type[0]}.")

    # get damage bins from fileFIXME
    static_path = os.path.join(run_dir, 'static')
    damage_bins = get_damage_bins(static_path=static_path)

    # maximum number of damage bins (individual items can have up to `total_bins` bins)
    if damage_bins.shape[0] == 0:
        max_Nbins = 1000
    else:
        max_Nbins = damage_bins.shape[0]

    # maximum number of entries is buff_size divided by the minimum entry size
    # (corresponding to a 1-bin only cdf)
    min_size_cdf_entry = damagecdfrec_stream.size + ProbMean.size + 4

    # each record from getmodel stream is expected to contain:
    # 1 damagecdfrec_stream obj, 1 int (Nbins), a number `Nbins` of ProbMean objects

    # use a memoryview of size twice the buff_size to ensure `read_into1` always reads the maximum amount
    # data possible (the largest between the pipe limit and the remaining memory to fill the memoryview)
    mv = memoryview(bytearray(buff_size * 2))
    int32_mv = np.ndarray(buff_size * 2 // 4, buffer=mv, dtype='i4')

    cursor = 0
    valid_buf = 0
    last_event_id = -1
    damagecdfs, Nbinss, recs = [], [], []
    len_read = 1
    size_cdf_entry = min_size_cdf_entry
    while True:
        if len_read > 0:
            # read stream from valid_buf onwards
            len_read = stream_in.readinto1(mv[valid_buf:])
            # extend the valid_buf by the same amount of data that was read
            valid_buf += len_read

        # read the streamed data into formatted data
        cursor, i, yield_event, event_id, damagecdf, Nbins, rec, last_event_id = stream_to_data(
            int32_mv, valid_buf, size_cdf_entry, max_Nbins, last_event_id)

        if yield_event:
            # event is fully read, append the last chunk of data to the list of this event
            damagecdfs.append(damagecdf[:i])
            Nbinss.append(Nbins[:i])
            recs.append(rec[:i])

            yield last_event_id, np.concatenate(damagecdfs), np.concatenate(Nbinss), np.concatenate(recs)

            # start a new list for the new event, storing the first element
            damagecdfs, Nbinss, recs = [damagecdf[i:i + 1]], [Nbins[i:i + 1]], [rec[i:i + 1]]
            last_event_id = event_id

        else:
            # the current event is not finished, keep appending data about this event
            damagecdfs.append(damagecdf[:i + 1])
            Nbinss.append(Nbins[:i + 1])
            recs.append(rec[:i + 1])

        # convert cursor to bytes
        cursor_buf = cursor * int32_mv.itemsize

        if valid_buf == cursor_buf:
            # this is the last cycle, all data has been read, append the last chunk of data
            damagecdfs.append(damagecdf[:i])
            Nbinss.append(Nbins[:i])
            recs.append(rec[:i])

            yield last_event_id, np.concatenate(damagecdfs), np.concatenate(Nbinss), np.concatenate(recs)
            break

        else:
            # this is not the last cycle
            # move the un-read data to the beginning of the memoryview
            mv[:valid_buf - cursor_buf] = mv[cursor_buf:valid_buf]

            # update the length of the valid data
            valid_buf -= cursor_buf


@njit(cache=True, fastmath=True)
def stream_to_data(int32_mv, valid_buf, size_cdf_entry, max_Nbins, last_event_id):
    """Parse streamed data into data arrays.

    Args:
        int32_mv (ndarray): int32 view of the buffer
        valid_buf (int): number of bytes with valid data
        size_cdf_entry (int): size (in bytes) of a single record
        max_Nbins (int): Maximum number of probability bins
        last_event_id (int): event_id of the last event that was completed

    Returns:
        int, int, bool, int, numpy.array[damagecdf], numpy.array[int32], numpy.array[ProbMean], int:
          number of int numbers read from the int32_mv ndarray, number of cdf data entries read,
          whether the current event (id=`event_id`) has been fully read, id of the event being read,
          chunk of damagecdf entries, chunk of Nbins entries, chunk of cdf record entries,
          event_id of the last event that was completed
    """
    valid_len = valid_buf // size_cdf_entry
    yield_event = False

    Nbins = np.zeros(valid_len, dtype='i4')
    damagecdf = np.zeros(valid_len, dtype=damagecdfrec)
    rec = np.zeros((valid_len, max_Nbins), dtype=ProbMean)

    i = 0
    cursor = 0
    while cursor < valid_len:
        event_id, cursor = int32_mv[cursor], cursor + 1
        damagecdf[i]['areaperil_id'], cursor = int32_mv[cursor:cursor + 1].view(areaperil_int)[0], cursor + 1
        damagecdf[i]['vulnerability_id'], cursor = int32_mv[cursor], cursor + 1
        Nbins[i] = int32_mv[cursor]
        cursor += 1

        for j in range(Nbins[i]):
            rec[i, j]['prob_to'] = int32_mv[cursor: cursor + oasis_float_to_int32_size].view(oasis_float)[0]
            cursor += oasis_float_to_int32_size
            rec[i, j]['bin_mean'] = int32_mv[cursor: cursor + oasis_float_to_int32_size].view(oasis_float)[0]
            cursor += oasis_float_to_int32_size

        if event_id != last_event_id:
            # a new event has started
            if last_event_id > 0:
                # if this is not the beginning of the very first event, yield the event we just completed
                yield_event = True

                return cursor, i, yield_event, event_id, damagecdf, Nbins, rec, last_event_id

            last_event_id = event_id

        i += 1

    return cursor, i - 1, yield_event, event_id, damagecdf, Nbins, rec, last_event_id


class LossWriter():
    # TODO clean this up and remove unused features
    def __init__(self, lossout, len_sample, buff_size=65536) -> None:

        # number of bytes to read at a given time.
        # number_size = 8 works only if loss in gulSampleslevelRec is float32.
        self.number_size = max(gulSampleslevelHeader.size, gulSampleslevelRec.size)  # bytes

        self.len_sample = len_sample
        self.lossout = lossout
        self.buff_size = buff_size  # bytes

        # compute how many numbers of size `number_size` fit in the buffer
        # for safety, take 1000 less than the compute number to allow flushing the buffer not too often
        # if -1 instead of -1000 is taken, it requires checking whether to flush or not for every write to mv.
        self.buff_safety = self.number_size * 1000
        self.nb_number = (self.buff_size + self.buff_safety) // self.number_size
        self.flush_number = self.nb_number - 4

        # define the raw memory view, the int32 view of it, and their respective cursors
        mv_size_bytes = buff_size * 2
        # mv_size_bytes = buff_size + self.number_size * 10  # doesn't work, produces wrong output (zeroes)
        self.mv = memoryview(bytearray(mv_size_bytes))
        self.int32_mv = np.ndarray(mv_size_bytes // self.number_size, buffer=self.mv, dtype='i4')
        # cannot use because the header is int int
        # self.loss_mv = np.ndarray(self.nb_number, buffer=self.mv, dtype=gulSampleslevelRec.dtype)
        # cannot use two views loss_mv and header_mv because it only works if oasis_float is float32.
        # if oasis_float is set to float64, the cursor will not map correctly both mv.
        self.cursor_bytes = 0
        self.cursor = 0

        # size(oasis_float)/size(i4)
        # TODO find a way to do that programmatically and test if this works with oasis_float=float64
        self.oasis_float_to_int32_size = 1

    def flush(self):
        # print("FLUSHING ", self.cursor, " ", self.cursor_bytes)
        self.lossout.write(self.mv[:self.cursor_bytes])
        self.cursor_bytes = 0
        self.cursor = 0
        # print("FLUSHED ", self.cursor, " ", self.cursor_bytes)

    def write_sample_header(self, event_id, item_id):
        self.int32_mv[self.cursor] = event_id
        self.cursor += 1
        self.int32_mv[self.cursor] = item_id
        self.cursor += 1
        self.cursor_bytes += gulSampleslevelHeader.size

    def write_sample_rec(self, sidx, loss):

        if self.cursor >= self.flush_number:
            self.flush()

        self.int32_mv[self.cursor] = sidx
        self.cursor += 1
        self.int32_mv[self.cursor:self.cursor + self.oasis_float_to_int32_size].view(oasis_float)[:] = loss
        self.cursor += self.oasis_float_to_int32_size
        self.cursor_bytes += gulSampleslevelRec.size

@njit(cache=True, fastmath=True)
def write_sample_header(event_id, item_id, int32_mv, cursor, cursor_bytes):
    """Write to buffer the header for the samples of this (event, item).

    Args:
        event_id (int32): event id.
        item_id (int32): item id.
        int32_mv (numpy.ndarray): int32 view of the memoryview where the output is buffered.
        cursor (int): index of int32_mv where to start writing.
        cursor_bytes (int): number of bytes written in int32_mv.

    Returns:
        int, int: updated values of cursor and cursor_bytes
    """
    int32_mv[cursor], cursor = event_id, cursor + 1
    int32_mv[cursor], cursor = item_id, cursor + 1
    cursor_bytes += gulSampleslevelHeader_size

    return cursor, cursor_bytes


@njit(cache=True, fastmath=True)
def write_sample_rec(sidx, loss, int32_mv, cursor, cursor_bytes):
    """Write to buffer a (sidx, loss) sample record.

    Args:
        sidx (int32): sidx number.
        loss (oasis_float): loss.
        int32_mv (numpy.ndarray): int32 view of the memoryview where the output is buffered.
        cursor (int): index of int32_mv where to start writing.
        cursor_bytes (int): number of bytes written in int32_mv.

    Returns:
        int, int: updated values of cursor and cursor_bytes
    """
    int32_mv[cursor], cursor = sidx, cursor + 1
    int32_mv[cursor:cursor + oasis_float_to_int32_size].view(oasis_float)[:] = loss
    cursor += oasis_float_to_int32_size
    cursor_bytes += gulSampleslevelRec_size

    return cursor, cursor_bytes


@njit(cache=True, fastmath=True)
def write_negative_sidx(max_loss_idx, max_loss, chance_of_loss_idx, chance_of_loss,
                        tiv_idx, tiv, std_dev_idx, std_dev, mean_idx, gul_mean,
                        int32_mv, cursor, cursor_bytes):
    """Write to buffer the negative sidx samples.

    Args:
        max_loss_idx (int32): max_loss_idx sidx number.
        max_loss (oasis_float): max_loss.
        chance_of_loss_idx (int32): chance_of_loss_idx sidx number.
        chance_of_loss_idx (oasis_float): chance_of_loss
        tiv_idx (int32): tiv_idx sidx number.
        tiv (oasis_float): tiv.
        std_dev_idx (int32): std_dev_idx sidx number.
        std_dev (oasis_float): std_dev.
        mean_idx (int32): mean_idx sidx number.
        gul_mean (oasis_float): gul_mean.
        int32_mv (numpy.ndarray): int32 view of the memoryview where the output is buffered.
        cursor (int): index of int32_mv where to start writing.
        cursor_bytes (int): number of bytes written in int32_mv.

    Returns:
        int, int: updated values of cursor and cursor_bytes
    """
    int32_mv[cursor], cursor = max_loss_idx, cursor + 1
    int32_mv[cursor:cursor + oasis_float_to_int32_size].view(oasis_float)[:] = max_loss
    cursor += oasis_float_to_int32_size
    cursor_bytes += gulSampleslevelRec_size

    int32_mv[cursor], cursor = chance_of_loss_idx, cursor + 1
    int32_mv[cursor:cursor + oasis_float_to_int32_size].view(oasis_float)[:] = chance_of_loss
    cursor += oasis_float_to_int32_size
    cursor_bytes += gulSampleslevelRec_size

    int32_mv[cursor], cursor = tiv_idx, cursor + 1
    int32_mv[cursor:cursor + oasis_float_to_int32_size].view(oasis_float)[:] = tiv
    cursor += oasis_float_to_int32_size
    cursor_bytes += gulSampleslevelRec_size

    int32_mv[cursor], cursor = std_dev_idx, cursor + 1
    int32_mv[cursor:cursor + oasis_float_to_int32_size].view(oasis_float)[:] = std_dev
    cursor += oasis_float_to_int32_size
    cursor_bytes += gulSampleslevelRec_size

    int32_mv[cursor], cursor = mean_idx, cursor + 1
    int32_mv[cursor:cursor + oasis_float_to_int32_size].view(oasis_float)[:] = gul_mean
    cursor += oasis_float_to_int32_size
    cursor_bytes += gulSampleslevelRec_size

    return cursor, cursor_bytes



"""
This file defines the data types that are loaded from the data files.

"""
import numpy as np
import numba as nb

from oasislmf.pytools.getmodel.common import oasis_float, areaperil_int

ITEM_ID_TYPE = nb.types.int32
ITEMS_DATA_MAP_TYPE = nb.types.UniTuple(nb.types.int64, 2)
COVERAGE_ID_TYPE = nb.types.int32

# negative sidx (definition)
MEAN_IDX = -1
STD_DEV_IDX = -2
TIV_IDX = -3
CHANCE_OF_LOSS_IDX = -4
MAX_LOSS_IDX = -5

NUM_IDX = 5

# negative sidx + NUM_IDX
SHIFTED_MEAN_IDX = MEAN_IDX + NUM_IDX
SHIFTED_STD_DEV_IDX = STD_DEV_IDX + NUM_IDX
SHIFTED_TIV_IDX = TIV_IDX + NUM_IDX
SHIFTED_CHANCE_OF_LOSS_IDX = CHANCE_OF_LOSS_IDX + NUM_IDX
SHIFTED_MAX_LOSS_IDX = MAX_LOSS_IDX + NUM_IDX


ITEM_MAP_KEY_TYPE = nb.types.Tuple((nb.types.uint32, nb.types.int32))
ITEM_MAP_VALUE_TYPE = nb.types.UniTuple(nb.types.int32, 3)

# compute the relative size of oasis_float vs int32
oasis_float_to_int32_size = oasis_float.itemsize // np.int32().itemsize

ProbMean = nb.from_dtype(np.dtype([('prob_to', oasis_float),
                                   ('bin_mean', oasis_float)
                                   ]))

damagecdfrec_stream = nb.from_dtype(np.dtype([('event_id', np.int32),
                                              ('areaperil_id', areaperil_int),
                                              ('vulnerability_id', np.int32)
                                              ]))

damagecdfrec = nb.from_dtype(np.dtype([('areaperil_id', areaperil_int),
                                       ('vulnerability_id', np.int32)
                                       ]))


gulSampleslevelHeader = nb.from_dtype(np.dtype([('event_id', 'i4'),
                                                ('item_id', 'i4'),
                                                ]))
gulSampleslevelHeader_size = gulSampleslevelHeader.size

gulSampleslevelRec = nb.from_dtype(np.dtype([('sidx', 'i4'),
                                             ('loss', oasis_float),
                                             ]))
gulSampleslevelRec_size = gulSampleslevelRec.size



"""
This file defines the data types that are loaded from the data files.
"""
import os

import numba as nb
import numpy as np

# filenames
footprint_filename = 'footprint.bin'
footprint_index_filename = 'footprint.idx'
zfootprint_filename = 'footprint.bin.z'
zfootprint_index_filename = 'footprint.idx.z'
csvfootprint_filename = 'footprint.csv'
parquetfootprint_filename = "footprint.parquet"
parquetfootprint_meta_filename = "footprint_parquet_meta.json"

areaperil_int = np.dtype(os.environ.get('AREAPERIL_TYPE', 'u4'))
oasis_float = np.dtype(os.environ.get('OASIS_FLOAT', 'f4'))

FootprintHeader = nb.from_dtype(np.dtype([('num_intensity_bins', np.int32),
                                          ('has_intensity_uncertainty', np.int32)
                                          ]))

Event = nb.from_dtype(np.dtype([('areaperil_id', areaperil_int),
                                ('intensity_bin_id', np.int32),
                                ('probability', oasis_float)
                                ]))

EventCSV =  nb.from_dtype(np.dtype([('event_id', np.int32),
                                    ('areaperil_id', areaperil_int),
                                    ('intensity_bin_id', np.int32),
                                    ('probability', oasis_float)
                                    ]))

EventIndexBin = nb.from_dtype(np.dtype([('event_id', np.int32),
                                        ('offset', np.int64),
                                        ('size', np.int64)
                                        ]))

EventIndexBinZ = nb.from_dtype(np.dtype([('event_id', np.int32),
                                         ('offset', np.int64),
                                         ('size', np.int64),
                                         ('d_size', np.int64)
                                         ]))

Index_type = nb.from_dtype(np.dtype([('start', np.int64),
                                     ('end', np.int64)
                                     ]))

Vulnerability = nb.from_dtype(np.dtype([('vulnerability_id', np.int32),
                                        ('intensity_bin_id', np.int32),
                                        ('damage_bin_id', np.int32),
                                        ('probability', oasis_float)
                                        ]))

Item = nb.from_dtype(np.dtype([('id', np.int32),
                               ('coverage_id', np.int32),
                               ('areaperil_id', areaperil_int),
                               ('vulnerability_id', np.int32),
                               ('group_id', np.int32)
                               ]))



import argparse
import atexit
import datetime
import logging
import math
import os
import pickle
import socket
from contextlib import ExitStack
from enum import Enum
from typing import Optional, Set, Tuple, List

import numpy as np

from oasislmf.pytools.getmodel.footprint import Footprint

# configuring process meta data
logging.basicConfig(
    filename='footprint_tcp_server.log',
    filemode='w',
    format='%(name)s - %(levelname)s - %(message)s',
    level=os.environ.get("LOGLEVEL", "INFO")
)
POINTER_PATH = str(os.path.dirname(os.path.realpath(__file__))) + "/pointer_flag.txt"
TCP_IP = '127.0.0.1'
TCP_PORT = 8080
PROCESSES_SUPPORTED = 100

from random import randint
MODEL_LOG_PATH = str(os.getcwd()) + f"/{randint(1,900)}_model_log.txt"


class OperationEnum(Enum):
    """
    Defines the different types of operations supported via bytes. To be passed through TCP port first to tell the
    server what type of operation is required.
    """
    GET_DATA = (1).to_bytes(4, byteorder='big')
    GET_NUM_INTENSITY_BINS = (2).to_bytes(4, byteorder='big')
    REGISTER = (3).to_bytes(4, byteorder='big')
    UNREGISTER = (4).to_bytes(4, byteorder='big')
    POLL_DATA = (5).to_bytes(4, byteorder='big')


class FootprintLayer:
    """
    This class is responsible for accessing the footprint data via TCP ports.

    Attributes:
        static_path (str): path to the static file to load the data
        ignore_file_type (Set[str]): collection of file types to ignore when loading
        file_data (Optional[Footprint]): footprint object to load
        socket (Optional[socket.socket]): the TCP socket in which data is sent
        count (int): the number of processes currently relying on the server
        total_expected (int): the total number of reliant processes expected
        total_served (int): the total number of processes that have ever registered through the server's lifetime
    """
    def __init__(self, static_path: str, total_expected: int, ignore_file_type: Set[str] = set()) -> None:
        """
        The constructor for the FootprintLayer class.

        Args:
            static_path: (str) path to the static file to load the data
            ignore_file_type: (Set[str]) collection of file types to ignore when loading
            total_expected: (int) the total number of reliant processes expected

        """
        self.static_path: str = static_path
        self.ignore_file_type: Set[str] = ignore_file_type
        self.file_data: Optional[Footprint] = None
        self.socket: Optional[socket.socket] = None
        self.count: int = 0
        self.total_expected: int = total_expected
        self.total_served: int = 0
        self._define_socket()

    def _define_socket(self) -> None:
        """
        Defines the self.socket attribute to the port and localhost.

        Returns: None
        """
        logging.info(f"defining socket for TCP server: {datetime.datetime.now()}")
        self.socket = socket.socket(socket.AF_INET, socket.SOCK_STREAM)
        server_address = (TCP_IP, TCP_PORT)
        self.socket.bind(server_address)
        self.socket.listen(PROCESSES_SUPPORTED)

    def _establish_shutdown_procedure(self) -> None:
        """
        Establishes the steps for shutdown, writing the pointer, and making sure that the pointer will be deleted
        and the self.socket is shutdown once the process running the server is exited.

        Returns: None
        """
        logging.info(f"establishing shutdown procedure: {datetime.datetime.now()}")
        # atexit.register(_shutdown_port, self.socket)
        pass

    @staticmethod
    def _stream_footprint_data(event_data: np.array, connection: socket.socket, event_id: int) -> None:
        """
        Serialises data then splits it into chunks of 500 in turn streaming through a connection.

        Args:
            event_data: (np.array) the data to be serialised and streamed through a connection
            connection: (socket.socket) the connection that the data is going to be streamed through

        Returns: None
        """
        raw_data: bytes = pickle.dumps(event_data)

        raw_data_buffer: List[bytes] = [raw_data[i:i + 60000] for i in range(0, len(raw_data), 60000)]

        for chunk in raw_data_buffer:
            connection.sendall(chunk)

    @staticmethod
    def _extract_header(header_data: bytes) -> Tuple[OperationEnum, Optional[int]]:
        """
        Extracts the operation type and event_id from the header.

        Args:
            header_data: (bytes) header data sent from the client and recieved via TCP

        Returns: (Tuple[OperationEnum, Optional[int]]) event_id is None if the Operation is not GET_DATA
        """
        operation_number_data = header_data[:4]
        event_id_data = header_data[4:12]

        operation: OperationEnum = OperationEnum(operation_number_data)

        if operation == OperationEnum.GET_DATA:
            event_id: Optional[int] = int.from_bytes(event_id_data, 'big')
        else:
            event_id: Optional[int] = None

        return operation, event_id

    def listen(self) -> None:
        """
        Fires off the process with an event loop serving footprint data.

        Returns: None
        """
        self._establish_shutdown_procedure()

        with ExitStack() as stack:
            footprint_obj = stack.enter_context(Footprint.load(static_path=self.static_path,
                                                               ignore_file_type=self.ignore_file_type))
            self.file_data = footprint_obj
            while True:
                try:
                    connection, client_address = self.socket.accept()
                    data = connection.recv(16)

                    if data:
                        operation, event_id = self._extract_header(header_data=data)

                        if operation == OperationEnum.GET_DATA:
                            event_data = self.file_data.get_event(event_id=event_id)

                            if event_id in self.file_data.footprint_index:
                                logging.info(f'event_id "{event_id}" retrieved from footprint index')
                                del self.file_data.footprint_index[event_id]
                            else:
                                logging.error(f'event_id "{event_id}" not in footprint_index')

                            FootprintLayer._stream_footprint_data(event_data=event_data, connection=connection, event_id=event_id)

                        elif operation == OperationEnum.GET_NUM_INTENSITY_BINS:

                            number_of_intensity_bins = self.file_data.num_intensity_bins
                            connection.sendall(number_of_intensity_bins.to_bytes(8, byteorder='big'))

                        elif operation == OperationEnum.REGISTER:
                            self.count += 1
                            self.total_served += 1
                            logging.info(f"connection registered: {self.count} for {client_address} {datetime.datetime.now()}")

                        elif operation == OperationEnum.UNREGISTER:
                            self.count -= 1
                            logging.info(f"connection unregistered: {self.count} for {client_address} {datetime.datetime.now()}")
                            if self.count <= 0 and self.total_expected == self.total_served:
                                logging.info(f"breaking event loop: {datetime.datetime.now()}")
                                self.socket.shutdown(socket.SHUT_RDWR)
                                break
                        connection.close()
                # Catch all errors, send to logger and keep running        
                except Exception as e:
                    logging.error(e)
            connection.close()


class FootprintLayerClient:
    """
    This class is responsible for connecting to the footprint server via TCP.
    """
    @classmethod
    def poll(cls) -> bool:
        """
        Checks to see if data server is running.

        Returns: (bool)
        """
        try:
            _ = cls._get_socket()
            return True
        except ConnectionRefusedError as e:
            logging.error('Failed to find server: {}'.format(e))
            return False

    @classmethod
    def _get_socket(cls) -> socket.socket:
        """
        Gets the socket using the cls.TCP_IP and cls.TCP_PORT.

        Returns: (socket.socket) a connected socket
        """
        current_socket = socket.socket(socket.AF_INET, socket.SOCK_STREAM)
        current_socket.connect((TCP_IP, TCP_PORT))
        current_socket.settimeout(10)
        return current_socket

    @classmethod
    def _define_shutdown_procedure(cls) -> None:
        """
        Unregisters the client to the server on exit of the process.

        Returns: None
        """
        atexit.register(cls.unregister)

    @classmethod
    def register(cls) -> None:
        """
        Registers the client with the server.

        Returns: None
        """
        connection_viable: bool = False
        while connection_viable is False:
            connection_viable = FootprintLayerClient.poll()

        current_socket = cls._get_socket()
        data: bytes = OperationEnum.REGISTER.value
        current_socket.sendall(data)
        current_socket.close()
        # cls._define_shutdown_procedure()

    @classmethod
    def unregister(cls) -> None:
        """
        Unregisters the client with the data server.

        Returns: None
        """
        current_socket = cls._get_socket()
        data: bytes = OperationEnum.UNREGISTER.value
        current_socket.sendall(data)
        current_socket.close()
        atexit.unregister(cls.unregister)

    @classmethod
    def get_number_of_intensity_bins(cls) -> int:
        """
        Gets the number of intensity bins from the footprint data.

        Returns: (int) the number of intensity bins
        """
        current_socket = cls._get_socket()

        data: bytes = OperationEnum.GET_NUM_INTENSITY_BINS.value
        current_socket.sendall(data)
        intensity_bins_data = current_socket.recv(8)
        current_socket.close()
        return int.from_bytes(intensity_bins_data, 'big')

    @classmethod
    def get_event(cls, event_id: int) -> np.array:
        """
        Gets the footprint data from the footprint based off the event_id.

        Args:
            event_id: (int) the event ID of the data required

        Returns: (np.array) footprint data based off the event_id
        """
        current_socket = cls._get_socket()

        data: bytes = OperationEnum.GET_DATA.value + int(event_id).to_bytes(8, byteorder='big')
        current_socket.sendall(data)

        raw_data_buffer: List[bytes] = []

        while True:
            raw_data = current_socket.recv(6000)
            if not raw_data:
                break
            raw_data_buffer.append(raw_data)

        return pickle.loads(b"".join(raw_data_buffer))


def _shutdown_port(connection: socket.socket) -> None:
    logging.info(f"socket is shutting down: {datetime.datetime.now()}")
    connection.shutdown(socket.SHUT_RDWR)


def main() -> None:
    parser = argparse.ArgumentParser()
    parser.add_argument("p", help="path to static file", type=str)
    parser.add_argument("n", help="number of processes expected to be reliant on server", type=int)
    args = parser.parse_args()
    server = FootprintLayer(static_path=args.p, total_expected=args.n)
    server.listen()


if __name__ == "__main__":
    main()




"""
    Python utilities used for setting up the structure of the run directory
    in which to prepare the inputs to run a model or generate deterministic
    losses, and store the outputs.
"""
__all__ = [
    'check_binary_tar_file',
    'check_conversion_tools',
    'check_inputs_directory',
    'cleanup_bin_directory',
    'create_binary_tar_file',
    'csv_to_bin',
    'prepare_run_directory',
    'prepare_run_inputs'
]


import errno
import csv
import filecmp
import glob
import logging
import os
import re
import shutil
import shutilwhich
import subprocess
import tarfile
import pandas as pd

from itertools import chain

from pathlib import Path

from ..utils.exceptions import OasisException
from ..utils.log import oasis_log
from ..utils.defaults import STATIC_DATA_FP
from .files import TAR_FILE, INPUT_FILES, GUL_INPUT_FILES, IL_INPUT_FILES
from .bash import leccalc_enabled, ord_enabled, ORD_LECCALC


@oasis_log
def prepare_run_directory(
    run_dir,
    oasis_src_fp,
    model_data_fp,
    analysis_settings_fp,
    inputs_archive=None,
    user_data_dir=None,
    ri=False,
):
    """
    Ensures that the model run directory has the correct folder structure in
    order for the model run script (ktools) to be executed. Without the RI
    flag the model run directory will have the following structure

    ::

        <run_directory>
        |-- fifo/
        |-- input/
        |   `-- csv/
        |-- output/
        |-- static/
        |-- work/
        |-- analysis_settings.json
        `-- run_ktools.sh


    where the direct GUL and/or FM input files exist in the ``input/csv``
    subfolder and the corresponding binaries exist in the ``input`` subfolder.

    With the RI flag the model run directory has the following structure

    ::
        <run_directory>
        |-- fifo
        |-- input
        |-- RI_1
        |-- RI_2
        |-- ...
        |-- output
        |-- static
        |-- work
        |-- ri_layers.json
        |-- analysis_settings.json
        `-- run_ktools.sh

    where the direct GUL and/or FM input files, and the corresponding binaries
    exist in the ``input`` subfolder, and the RI layer input files and binaries
    exist in the ``RI`` prefixed subfolders.

    If any subfolders are missing they are created.

    Optionally, if the path to a set of Oasis files is provided then they
    are copied into the ``input/csv`` subfolder.

    Optionally, if the path to the analysis settings JSON file is provided
    then it is copied to the base of the run directory.

    Optionally, if the path to model data is provided then the files are
    symlinked into the ``static`` subfolder provided the OS is of type
    Darwin or Linux, otherwise the source folder tree is recursively
    copied into the ``static`` subfolder.

    :param run_dir: the model run directory
    :type run_dir: str

    :param oasis_src_fp: path to a set of Oasis files
    :type oasis_src_fp: str

    :param ri: Boolean flag for RI mode
    :type ri: bool

    :param analysis_settings_fp: analysis settings JSON file path
    :type analysis_settings_fp: str

    :param model_data_fp: model data source path
    :type model_data_fp: str

    :param inputs_archive: path to a tar file containing input files
    :type inputs_archive: str

    :param: user_data_dir: path to a directory containing additional user-supplied model data
    :type user_data_dir: str
    """
    try:
        for subdir in ['fifo', 'output', 'static', 'work']:
            Path(run_dir, subdir).mkdir(parents=True, exist_ok=True)

        if not inputs_archive:
            Path(run_dir, 'input').mkdir(parents=True, exist_ok=True)
        else:
            with tarfile.open(inputs_archive) as input_tarfile:
                p = os.path.join(run_dir, 'input') if not ri else os.path.join(run_dir)
                input_tarfile.extractall(path=p)

        oasis_dst_fp = os.path.join(run_dir, 'input')

        for p in os.listdir(oasis_src_fp):
            src = os.path.join(oasis_src_fp, p)
            if src.endswith('.tar') or src.endswith('.tar.gz'):
                continue
            dst = os.path.join(oasis_dst_fp, p)
            if not (re.match(r'RI_\d+$', p) or p == 'ri_layers.json'):
                shutil.copy2(src, oasis_dst_fp) if not (os.path.exists(dst) and filecmp.cmp(src, dst)) else None
            else:
                shutil.move(src, run_dir)

        dst = os.path.join(run_dir, 'analysis_settings.json')
        shutil.copy(analysis_settings_fp, dst) if not (os.path.exists(dst) and filecmp.cmp(analysis_settings_fp, dst, shallow=False)) else None

        model_data_dst_fp = os.path.join(run_dir, 'static')

        try:
            for sourcefile in glob.glob(os.path.join(model_data_fp, '*')):
                destfile = os.path.join(model_data_dst_fp, os.path.basename(sourcefile))

                if os.name == 'nt':
                    shutil.copy(sourcefile, destfile)
                else:
                    os.symlink(sourcefile, destfile)
        except OSError as e:
            if not (e.errno == errno.EEXIST and os.path.islink(destfile) and os.name != 'nt'):
                raise e
            else:
                # If the link already exists, check files are different replace it
                if os.readlink(destfile) != os.path.abspath(sourcefile):
                    os.symlink(sourcefile, destfile + ".tmp")
                    os.replace(destfile + ".tmp", destfile)

        if user_data_dir and os.path.exists(user_data_dir):
            for sourcefile in glob.glob(os.path.join(user_data_dir, '*')):
                destfile = os.path.join(model_data_dst_fp, os.path.basename(sourcefile))

                try:
                    if os.name == 'nt':
                        shutil.copy(sourcefile, destfile)
                    else:
                        os.symlink(sourcefile, destfile)
                except OSError as e:
                    if not (e.errno == errno.EEXIST and os.path.islink(destfile) and os.name != 'nt'):
                        raise e
                    else:
                        # If the link already exists, check files are different replace it
                        if os.readlink(destfile) != os.path.abspath(sourcefile):
                            os.symlink(sourcefile, destfile + ".tmp")
                            os.replace(destfile + ".tmp", destfile)

    except OSError as e:
        raise OasisException("Error preparing the 'run' directory: {}".format(e))


def _create_return_period_bin(run_dir, return_periods):
    csv_fp = os.path.join(run_dir, 'input', 'returnperiods.csv')
    bin_fp = os.path.join(run_dir, 'input', 'returnperiods.bin')
    pd.DataFrame(
        return_periods,
        columns =['return_period']).sort_values(ascending=False, by=['return_period']
    ).to_csv(csv_fp, index=False)

    try:
        cmd_str = "returnperiodtobin < \"{}\" > \"{}\"".format(csv_fp, bin_fp)
        subprocess.check_call(cmd_str, stderr=subprocess.STDOUT, shell=True)
    except subprocess.CalledProcessError as e:
        raise OasisException("Error while converting returnperiods.csv to ktools binary format: {}".format(e))

def _create_events_bin(run_dir, event_ids):
    csv_fp = os.path.join(run_dir, 'input', 'events.csv')
    bin_fp = os.path.join(run_dir, 'input', 'events.bin')
    pd.DataFrame(
        event_ids,
        columns =['event_id']).sort_values(ascending=True, by=['event_id']
    ).to_csv(csv_fp, index=False)

    try:
        cmd_str = "evetobin < \"{}\" > \"{}\"".format(csv_fp, bin_fp)
        subprocess.check_call(cmd_str, stderr=subprocess.STDOUT, shell=True)
    except subprocess.CalledProcessError as e:
        raise OasisException("Error while converting events.csv to ktools binary format: {}".format(e))

def _create_quantile_bin(run_dir, quantiles):
    csv_fp = os.path.join(run_dir, 'input', 'quantile.csv')
    bin_fp = os.path.join(run_dir, 'input', 'quantile.bin')
    pd.DataFrame(
        quantiles,
        dtype='float',
        columns =['quantile']).sort_values(ascending=True, by=['quantile']
    ).to_csv(csv_fp, index=False)

    try:
        cmd_str = "quantiletobin < \"{}\" > \"{}\"".format(csv_fp, bin_fp)
        subprocess.check_call(cmd_str, stderr=subprocess.STDOUT, shell=True)
    except subprocess.CalledProcessError as e:
        raise OasisException("Error while converting quantile.csv to ktools binary format: {}".format(e))

def _load_default_quantile_bin(run_dir):
    default_quantile_fp = os.path.join(STATIC_DATA_FP, 'quantile.csv')
    _create_quantile_bin(run_dir,
        pd.read_csv(default_quantile_fp)["quantile"].tolist()
    )

def _prepare_input_bin(run_dir, bin_name, model_settings, setting_key=None, ri=False):
    bin_fp = os.path.join(run_dir, 'input', '{}.bin'.format(bin_name))
    if not os.path.exists(bin_fp):
        setting_val = model_settings.get(setting_key)

        if not setting_val:
            model_data_bin_fp = os.path.join(run_dir, 'static', '{}.bin'.format(bin_name))
        else:
            # 'verbatim' -  Try setting value as given
            model_data_bin_fp = os.path.join(run_dir, 'static', '{}_{}.bin'.format(bin_name, str(setting_val)))
            if not os.path.isfile(model_data_bin_fp):
                # 'compatibility' - Fallback name formatting to keep existing conversion
                setting_val = str(setting_val).replace(' ', '_').lower()
                model_data_bin_fp = os.path.join(run_dir, 'static', '{}_{}.bin'.format(bin_name, setting_val))

        if not os.path.exists(model_data_bin_fp):
            raise OasisException('Could not find {} data file: {}'.format(bin_name, model_data_bin_fp))

        shutil.copyfile(model_data_bin_fp, bin_fp)


def _calc_selected(analysis_settings, calc_type_list):
    """
    Return True, if any options in "calc_type_list" are set in the analysis settings file

    :param calc_type_list: List of string values or ktools outputs, e.g. `eltcalc`, `lec_output`, `aalcalc` or `pltcalc`
    :type  calc_type_list: list
    """
    gul_section = analysis_settings.get('gul_summaries')
    il_section = analysis_settings.get('il_summaries')
    ri_section = analysis_settings.get('ri_summaries')

    for calc_type in calc_type_list:

        if gul_section:
            if any(gul_summary.get(calc_type, None) or gul_summary.get('ord_output', {}).get(calc_type) for gul_summary in gul_section):
                return True
        if il_section:
            if any(il_summary.get(calc_type, None) or il_summary.get('ord_output', {}).get(calc_type) for il_summary in il_section):
                return True
        if ri_section:
            if any(ri_summary.get(calc_type, None) or ri_summary.get('ord_output', {}).get(calc_type) for ri_summary in ri_section):
                return True
    return False


def _leccalc_selected(analysis_settings):
    """ return True if either 'leccalc' or 'ordleccalc' is referenced in the analysis settings file
    """
    is_in_gul = False
    is_in_il = False
    is_in_ri = False

    gul_section = analysis_settings.get('gul_summaries')
    il_section = analysis_settings.get('il_summaries')
    ri_section = analysis_settings.get('ri_summaries')

    if gul_section:
        is_in_gul = any(leccalc_enabled(gul_summary) or ord_enabled(gul_summary, ORD_LECCALC) for gul_summary in gul_section)
    if il_section:
        is_in_il = any(leccalc_enabled(il_summary) or ord_enabled(il_summary, ORD_LECCALC) for il_summary in il_section)
    if ri_section:
        is_in_ri = any(leccalc_enabled(ri_summary) or ord_enabled(ri_summary, ORD_LECCALC) for ri_summary in ri_section)

    return any([is_in_gul, is_in_il, is_in_ri])


@oasis_log
def prepare_run_inputs(analysis_settings, run_dir, ri=False):
    """
    Sets up binary files in the model inputs directory.

    :param analysis_settings: model analysis settings dict
    :type analysis_settings: dict

    :param run_dir: model run directory
    :type run_dir: str
    """
    try:
        model_settings = analysis_settings.get('model_settings', {})
        # Prepare events.bin
        if analysis_settings.get('event_ids'):
            # Create events file from user input
            _create_events_bin(run_dir, analysis_settings.get('event_ids'))
        else:
            # copy selected event set from static
            _prepare_input_bin(run_dir, 'events', model_settings, setting_key='event_set', ri=ri)

        # Prepare quantile.bin
        if analysis_settings.get('quantiles'):
            # 1. Create quantile file from user input
            _create_quantile_bin(run_dir, analysis_settings.get('quantiles'))
        elif  _calc_selected(analysis_settings, ['plt_quantile', 'elt_quantile']):
            # 2. copy quantile file from model data
            if os.path.exists(os.path.join(run_dir, 'static', 'quantile.bin')):
                _prepare_input_bin(run_dir, 'quantile', model_settings, ri=ri)
            else:
            # 3. Create quantile file from package `_data/quantile.csv`
                _load_default_quantile_bin(run_dir)

        # Prepare occurrence / returnperiod depending on output calcs selected
        if _leccalc_selected(analysis_settings):
            if analysis_settings.get('return_periods'):
                # Create return periods from user input
                _create_return_period_bin(run_dir, analysis_settings.get('return_periods'))
            else:
                # copy return periods from static
                _prepare_input_bin(run_dir, 'returnperiods', model_settings)

            _prepare_input_bin(run_dir, 'occurrence', model_settings, setting_key='event_occurrence_id', ri=ri)
        elif _calc_selected(analysis_settings, [
            'pltcalc', 'aalcalc', 'alt_period', 'elt_moment', 'elt_quantile',
            'elt_sample', 'plt_moment', 'plt_quantile', 'plt_sample'
        ]):
            _prepare_input_bin(run_dir, 'occurrence', model_settings, setting_key='event_occurrence_id', ri=ri)

        # Prepare periods.bin
        if os.path.exists(os.path.join(run_dir, 'static', 'periods.bin')):
            _prepare_input_bin(run_dir, 'periods', model_settings, ri=ri)


    except (OSError, IOError) as e:
        raise OasisException("Error preparing the model 'inputs' directory: {}".format(e))


@oasis_log
def check_inputs_directory(directory_to_check, il=False, ri=False, check_binaries=True):
    """
    Check that all the required files are present in the directory.

    :param directory_to_check: directory containing the CSV files
    :type directory_to_check: string

    :param il: check insuured loss files
    :type il: bool

    :param il: check resinsurance sub-folders
    :type il: bool

    :param check_binaries: check binary files are not present
    :type check_binaries: bool
    """
    # Check the top level directory, that containes the core files and any direct FM files
    _check_each_inputs_directory(directory_to_check, il=il, check_binaries=check_binaries)

    if ri:
        for ri_directory_to_check in glob.glob('{}{}RI_\d+$'.format(directory_to_check, os.path.sep)):
            _check_each_inputs_directory(ri_directory_to_check, il=True, check_binaries=check_binaries)


def _check_each_inputs_directory(directory_to_check, il=False, check_binaries=True):
    """
    Detailed check of a specific directory
    """

    if il:
        input_files = (f['name'] for f in INPUT_FILES.values() if f['type'] != 'optional')
    else:
        input_files = (f['name'] for f in INPUT_FILES.values() if f['type'] not in ['optional', 'il'])

    for input_file in input_files:
        file_path = os.path.join(directory_to_check, input_file + ".csv")
        if not os.path.exists(file_path):
            raise OasisException("Failed to find {}".format(file_path))

        if check_binaries:
            file_path = os.path.join(directory_to_check, input_file + ".bin")
            if os.path.exists(file_path):
                raise OasisException("Binary file already exists: {}".format(file_path))


@oasis_log
def csv_to_bin(csv_directory, bin_directory, il=False, ri=False):
    """
    Create the binary files.

    :param csv_directory: the directory containing the CSV files
    :type csv_directory: str

    :param bin_directory: the directory to write the binary files
    :type bin_directory: str

    :param il: whether to create the binaries required for insured loss calculations
    :type il: bool

    :param ri: whether to create the binaries required for reinsurance calculations
    :type ri: bool

    :raises OasisException: If one of the conversions fails
    """
    csvdir = os.path.abspath(csv_directory)
    bindir = os.path.abspath(bin_directory)

    il = il or ri

    _csv_to_bin(csvdir, bindir, il)

    if ri:
        for ri_csvdir in glob.glob('{}{}RI_[0-9]*'.format(csvdir, os.sep)):
            _csv_to_bin(
                ri_csvdir, os.path.join(bindir, os.path.basename(ri_csvdir)), il=True)


def _csv_to_bin(csv_directory, bin_directory, il=False):
    """
    Create a set of binary files.
    """
    if not os.path.exists(bin_directory):
        os.mkdir(bin_directory)

    if il:
        input_files = INPUT_FILES.values()
    else:
        input_files = (f for f in INPUT_FILES.values() if f['type'] != 'il')

    for input_file in input_files:
        conversion_tool = input_file['conversion_tool']
        input_file_path = os.path.join(csv_directory, '{}.csv'.format(input_file['name']))
        if not os.path.exists(input_file_path):
            continue

        output_file_path = os.path.join(bin_directory, '{}.bin'.format(input_file['name']))

        # If input file is different for step policies, apply flag when
        # executing conversion tool should step policies be present
        step_flag = input_file.get('step_flag')
        col_names = []
        if step_flag:
            with open(input_file_path) as f:
                reader = csv.reader(f)
                col_names = next(reader)

        if 'step_id' in col_names:
            output_file_path = os.path.join(
                bin_directory, '{}{}.bin'.format(input_file['name'], '_step')
            )

            cmd_str = "{} {} < \"{}\" > \"{}\"".format(conversion_tool, step_flag, input_file_path, output_file_path)
        else:
            cmd_str = "{} < \"{}\" > \"{}\"".format(conversion_tool, input_file_path, output_file_path)

        try:
            subprocess.check_call(cmd_str, stderr=subprocess.STDOUT, shell=True)
        except subprocess.CalledProcessError as e:
            raise OasisException("Error while converting csv's to ktools binary format: {}".format(e))


@oasis_log
def check_binary_tar_file(tar_file_path, check_il=False):
    """
    Checks that all required files are present

    :param tar_file_path: Path to the tar file to check
    :type tar_file_path: str

    :param check_il: Flag whether to check insured loss files
    :type check_il: bool

    :raises OasisException: If a required file is missing

    :return: True if all required files are present, False otherwise
    :rtype: bool
    """
    expected_members = ('{}.bin'.format(f['name']) for f in GUL_INPUT_FILES.values())

    if check_il:
        expected_members = chain(expected_members, ('{}.bin'.format(f['name']) for f in IL_INPUT_FILES.values()))

    with tarfile.open(tar_file_path) as tar:
        for member in expected_members:
            try:
                tar.getmember(member)
            except KeyError:
                raise OasisException('{} is missing from the tar file {}.'.format(member, tar_file_path))

    return True


@oasis_log
def create_binary_tar_file(directory):
    """
    Package the binaries in a gzipped tar.

    :param directory: Path containing the binaries
    :type tar_file_path: str
    """
    with tarfile.open(os.path.join(directory, TAR_FILE), "w:gz") as tar:
        for f in glob.glob('{}*{}*.bin'.format(directory, os.sep)):
            logging.info("Adding {} {}".format(f, os.path.relpath(f, directory)))
            relpath = os.path.relpath(f, directory)
            tar.add(f, arcname=relpath)

        for f in glob.glob('{}*{}*{}*.bin'.format(directory, os.sep, os.sep)):
            relpath = os.path.relpath(f, directory)
            tar.add(f, arcname=relpath)


@oasis_log
def check_conversion_tools(il=False):
    """
    Check that the conversion tools are available

    :param il: Flag whether to check insured loss tools
    :type il: bool

    :return: True if all required tools are present, False otherwise
    :rtype: bool
    """
    if il:
        input_files = INPUT_FILES.values()
    else:
        input_files = (f for f in INPUT_FILES.values() if f['type'] != 'il')

    for input_file in input_files:
        tool = input_file['conversion_tool']
        if shutilwhich.which(tool) is None:
            error_message = "Failed to find conversion tool: {}".format(tool)
            logging.error(error_message)
            raise OasisException(error_message)

    return True


@oasis_log
def cleanup_bin_directory(directory):
    """
    Clean the tar and binary files.
    """
    for file in chain([TAR_FILE], (f + '.bin' for f in INPUT_FILES.keys())):
        file_path = os.path.join(directory, file)
        if os.path.exists(file_path):
            os.remove(file_path)



import hashlib
import json

import numpy as np
import pandas as pd

from oasislmf.pytools.getmodel.common import Item


class Singleton(type):

    _instances = {}

    def __call__(cls, *args, **kwargs):
        if cls not in cls._instances:
            cls._instances[cls] = super(Singleton, cls).__call__(*args, **kwargs)
        return cls._instances[cls]


class IdMap(dict, metaclass=Singleton):

    def __init__(self) -> None:
        super().__init__({})


def process_hashed_item(item_data: "np.array[Item]", total_len: int) -> "np.array[Item]":
    """
    Hash the group IDs of the items data.

    Args:
        item_data: (np.array[Item]) data to be processed from
        total_len: (int) the length of the item data being processed

    Returns: (np.array[Item]) Item data that has been hashed
    """
    buffer = np.zeros(total_len, dtype=Item)
    for x in range(total_len):
        buffer[x]["id"] = item_data[x]["id"]
        buffer[x]["coverage_id"] = item_data[x]["coverage_id"]
        buffer[x]["areaperil_id"] = item_data[x]["areaperil_id"]
        buffer[x]["vulnerability_id"] = item_data[x]["vulnerability_id"]
        buffer[x]["group_id"] = generate_group_id_hash(group_id=item_data[x]["group_id"])
    return buffer


def convert_item_file_ids_to_hash(input_directory: str) -> None:
    """
    Hashes the group ID in the items.bin file writing the result to that file.

    Args:
        input_directory: (str) the path to the input directory

    Returns: None
    """
    item_data = np.memmap(f"{input_directory}/items.bin", dtype=Item, mode='readwrite')

    for x in range(item_data.shape[0]):
        hashed_id = generate_group_id_hash(group_id=item_data[x]["group_id"])
        item_data[x]["group_id"] = hashed_id

    _write_hashed_id_map(input_directory=input_directory)


def convert_item_csv_to_hash(input_directory: str) -> None:
    """
    Rewrites the item CSV file to the same CSV file with hashed group IDs.

    Args:
        input_directory: (str) the directory of where the file is to be altered

    Returns: None
    """
    file_path: str = f"{input_directory}/items.csv"
    df = pd.read_csv(file_path)
    df['group_id'] = df['group_id'].apply(generate_group_id_hash)

    _write_hashed_id_map(input_directory=input_directory)
    df.to_csv(file_path, index=False)


def generate_group_id_hash(pre_hash_pointer: str) -> int:
    """
    Generates a hashed version of the group_id.

    Args:
        pre_hash_pointer: (int) the group ID to be hashed

    Returns: (int) the hashed group ID
    """
    id_map = IdMap()

    # with("./")

    # np.random.seed(int(group_id))
    hashed_id: int = id_map.get(
        pre_hash_pointer, int(hashlib.sha256(pre_hash_pointer.encode('utf-8')).hexdigest(), 16) % 10**8
    )
    id_map[pre_hash_pointer] = hashed_id
    return hashed_id


def _write_hashed_id_map(input_directory: str) -> None:
    """
    Writes the hashed ID maps to the JSON file.

    Args:
        input_directory: (str) the directory of where the map is going to be written

    Returns: None
    """
    with open(f"{input_directory}/group_id_map.json", "w") as file:
        id_map = IdMap()
        json_data = json.dumps(id_map)
        file.write(json_data)



import json
import os
from contextlib import ExitStack

import pandas as pd
import pyarrow as pa
import pyarrow.parquet as pq

from oasislmf.pytools.getmodel.footprint import Footprint


def convert_bin_to_parquet(static_path: str) -> None:
    """
    Converts the data from a binary file to a parquet file.

    Args:
        static_path: (str) the path to the static file

    Returns: None
    """
    with ExitStack() as stack:
        footprint_obj = stack.enter_context(Footprint.load(static_path=static_path,
                                                           ignore_file_type={'z', 'csv', 'parquet'}))
        index_data = footprint_obj.footprint_index

        meta_data = {
            "num_intensity_bins": footprint_obj.num_intensity_bins,
            "has_intensity_uncertainty": True if footprint_obj.has_intensity_uncertainty == 1 else False
        }

        for event_id in index_data.keys():
            data_slice = footprint_obj.get_event(event_id)
            df = pd.DataFrame(data_slice)
            df["event_id"] = event_id
            pq.write_to_dataset(
                pa.Table.from_pandas(df),
                root_path=f'{static_path}/footprint.parquet',
                partition_cols=['event_id'],
                compression="BROTLI"
            )
        with open(f'{static_path}/footprint_parquet_meta.json', 'w') as outfile:
            json.dump(meta_data, outfile)


def main() -> None:
    """
    The entrypoint for convertbintoparquet command converting the footprint.bin file in the current directory to a
    footprint.parquet file.

    Returns: None
    """
    convert_bin_to_parquet(static_path=str(os.getcwd()))



"""
This file houses the classes that load the footprint data from compressed, binary, and CSV files.
"""
import json
import logging
import mmap
import os
from contextlib import ExitStack
from typing import Dict, List, Union
from zlib import decompress

import numpy as np
import pandas as pd
import pyarrow.parquet as pq
import numba as nb

from .common import (FootprintHeader, EventIndexBin, EventIndexBinZ, Event, EventCSV,
                     footprint_filename, footprint_index_filename, zfootprint_filename, zfootprint_index_filename,
                     csvfootprint_filename, parquetfootprint_filename, parquetfootprint_meta_filename)

logger = logging.getLogger(__name__)

uncompressedMask = 1 << 1
intensityMask = 1


CURRENT_DIRECTORY = str(os.getcwd())


class OasisFootPrintError(Exception):
    """
    Raises exceptions when loading footprints.
    """
    def __init__(self, message: str) -> None:
        """
        The constructor of the OasisFootPrintError class.

        Args:
            message: (str) the message to be raised
        """
        super().__init__(message)


class Footprint:
    """
    This class is the base class for the footprint loaders.

    Attributes:
        static_path (str): the path to the static files directory
        stack (ExitStack): the context manager that combines other context managers and cleanup functions
    """
    def __init__(self, static_path) -> None:
        """
        The constructor for the Footprint class.

        Args:
            static_path: (str) the path to the static files directory
        """
        self.static_path = static_path
        self.stack = ExitStack()

    def __exit__(self, exc_type, exc_value, exc_traceback):
        self.stack.__exit__(exc_type, exc_value, exc_traceback)

    @classmethod
    def load(cls, static_path, ignore_file_type=set()):
        """
        Loads the loading classes defined in this file checking to see if the files are in the static path
        whilst doing so. The loading goes through the hierarchy with the following order:

        -> compressed binary file
        -> binary file
        -> CSV file

        If the compressed binary file is present, this will be loaded. If it is not, then the binary file will be loaded
        and so on.

        Args:
            static_path: (str) the path to the static files directory
            ignore_file_type: (Set[str]) type of file to be skipped in the hierarchy. This can be a choice of:

            parquet
            json
            z
            bin
            idx

        Returns: (Union[FootprintBinZ, FootprintBin, FootprintCsv]) the loaded class
        """
        priorities = [
            FootprintParquet,
            FootprintBinZ,
            FootprintBin,
            FootprintCsv
        ]

        for footprint_class in priorities:
            for filename in footprint_class.footprint_filenames:
                if (not os.path.exists(os.path.join(static_path, filename))
                        or filename.rsplit('.', 1)[-1] in ignore_file_type):
                    valid = False
                    break
            else:
                valid = True
            if valid:
                for filename in footprint_class.footprint_filenames:
                    logger.debug(f"loading {os.path.join(static_path, filename)}")
                return footprint_class(static_path)
        else:
            if os.path.isfile(f"{static_path}/footprint.parquet"):
                raise OasisFootPrintError(
                    message=f"footprint.parquet needs to be partitioned in order to work, please see: "
                            f"oasislmf.pytools.data_layer.conversions.footprint => convert_bin_to_parquet"
                )
            raise OasisFootPrintError(message=f"no valid footprint in {static_path}")

    def get_event(self, event_id):
        raise NotImplementedError()


class FootprintCsv(Footprint):
    """
    This class is responsible for loading footprint data from CSV.

    Attributes (when in context):
        footprint (np.array[EventCSV]): event data loaded from the CSV file
        num_intensity_bins (int): number of intensity bins in the data
        has_intensity_uncertainty (bool): if the data has uncertainty
        footprint_index (dict): map of footprint IDs with the index in the data
    """
    footprint_filenames = [csvfootprint_filename]

    def __enter__(self):
        self.footprint = np.genfromtxt(os.path.join(self.static_path, "footprint.csv"), dtype=EventCSV, delimiter=",")
        self.num_intensity_bins = max(self.footprint['intensity_bin_id'])

        footprint_df = pd.DataFrame(self.footprint, columns=self.footprint.dtype.names)
        self.has_intensity_uncertainty = footprint_df.groupby(['event_id','areaperil_id']).size().max() > 1

        footprint_index_df = footprint_df.groupby('event_id', as_index=False).size()
        footprint_index_df['offset'] = footprint_index_df['size'].cumsum() - footprint_index_df['size']
        footprint_index_df.set_index('event_id', inplace=True)
        self.footprint_index = footprint_index_df.to_dict('index')

        return self

    def get_event(self, event_id):
        """
        Gets the event from self.footprint based off the event ID passed in.

        Args:
            event_id: (int) the ID belonging to the Event being extracted

        Returns: (np.array[EventCSV]) the event that was extracted
        """
        event_info = self.footprint_index.get(event_id)
        if event_info is None:
            return
        else:
            return self.footprint[event_info['offset']: event_info['offset'] + event_info['size']]


class FootprintBin(Footprint):
    """
    This class is responsible loading the event data from the footprint binary files.

    Attributes (when in context):
        footprint (mmap.mmap): loaded data from the binary file which has header and then Event data
        num_intensity_bins (int): number of intensity bins in the data
        has_intensity_uncertainty (bool): if the data has uncertainty
        footprint_index (dict): map of footprint IDs with the index in the data
    """
    footprint_filenames = [footprint_filename, footprint_index_filename]

    def __enter__(self):
        self.footprint = self.stack.enter_context(
            mmap.mmap(os.open(os.path.join(self.static_path, footprint_filename), flags=os.O_RDONLY),
                      length=0, access=mmap.ACCESS_READ
                      )
        )
        footprint_header = np.frombuffer(bytearray(self.footprint[:FootprintHeader.size]), dtype=FootprintHeader)
        self.num_intensity_bins = int(footprint_header['num_intensity_bins'])
        self.has_intensity_uncertainty = int(footprint_header['has_intensity_uncertainty'] & intensityMask)

        footprint_mmap = np.memmap(
            os.path.join(self.static_path, footprint_index_filename),
            dtype=EventIndexBin,
            mode='r'
        )

        self.footprint_index = pd.DataFrame(
            footprint_mmap,
            columns=footprint_mmap.dtype.names
        ).set_index('event_id').to_dict('index')
        return self

    def get_event(self, event_id):
        """
        Gets the event from self.footprint based off the event ID passed in.

        Args:
            event_id: (int) the ID belonging to the Event being extracted

        Returns: (np.array(Event)) the event that was extracted
        """
        event_info = self.footprint_index.get(event_id)
        if event_info is None:
            return
        else:
            return np.frombuffer(self.footprint[event_info['offset']: event_info['offset'] + event_info['size']], Event)


class FootprintBinZ(Footprint):
    """
    This class is responsible for loading event data from compressed event data.

    Attributes (when in context):
        zfootprint (mmap.mmap): loaded data from the compressed binary file which has header and then Event data
        num_intensity_bins (int): number of intensity bins in the data
        has_intensity_uncertainty (bool): if the data has uncertainty
        uncompressed_size (int): the size in which the data is when it is decompressed
        index_dtype (Union[EventIndexBinZ, EventIndexBin]) the data type
        footprint_index (dict): map of footprint IDs with the index in the data
    """
    footprint_filenames = [zfootprint_filename, zfootprint_index_filename]

    def __enter__(self):
        self.zfootprint = self.stack.enter_context(mmap.mmap(os.open(os.path.join(self.static_path, zfootprint_filename), flags=os.O_RDONLY),
                                                             length=0, access=mmap.ACCESS_READ))

        footprint_header = np.frombuffer(bytearray(self.zfootprint[:FootprintHeader.size]), dtype=FootprintHeader)
        self.num_intensity_bins = int(footprint_header['num_intensity_bins'])
        self.has_intensity_uncertainty = int(footprint_header['has_intensity_uncertainty'] & intensityMask)
        self.uncompressed_size = int((footprint_header['has_intensity_uncertainty'] & uncompressedMask) >> 1)
        if self.uncompressed_size:
            self.index_dtype = EventIndexBinZ
        else:
            self.index_dtype = EventIndexBin

        zfootprint_mmap = np.memmap(os.path.join(self.static_path, zfootprint_index_filename), dtype=self.index_dtype, mode='r')
        self.footprint_index = pd.DataFrame(zfootprint_mmap, columns=zfootprint_mmap.dtype.names).set_index('event_id').to_dict('index')
        return self

    def get_event(self, event_id):
        """
        Gets the event from self.zfootprint based off the event ID passed in.

        Args:
            event_id: (int) the ID belonging to the Event being extracted

        Returns: (np.array[Event]) the event that was extracted
        """
        event_info = self.footprint_index.get(event_id)
        if event_info is None:
            return
        else:
            zdata = self.zfootprint[event_info['offset']: event_info['offset']+event_info['size']]
            data = decompress(zdata)
            return np.frombuffer(data, Event)


class FootprintParquet(Footprint):
    """
    This class is responsible for loading event data from parquet event data.

    Attributes (when in context):
        num_intensity_bins (int): number of intensity bins in the data
        has_intensity_uncertainty (bool): if the data has uncertainty
        footprint_index (dict): map of footprint IDs with the index in the data
    """
    footprint_filenames: List[str] = [parquetfootprint_filename, parquetfootprint_meta_filename]

    def __enter__(self):
        with open(f'{self.static_path}/{parquetfootprint_meta_filename}', 'r') as outfile:
            meta_data: Dict[str, Union[int, bool]] = json.load(outfile)

        self.num_intensity_bins = int(meta_data['num_intensity_bins'])
        self.has_intensity_uncertainty = int(meta_data['has_intensity_uncertainty'] & intensityMask)

        return self

    def get_event(self, event_id: int):
        """
        Gets the event data from the partitioned parquet data file.

        Args:
            event_id: (int) the ID belonging to the Event being extracted

        Returns: (np.array[Event]) the event that was extracted
        """
        try:
            handle = pq.ParquetDataset(f'./static/footprint.parquet/event_id={event_id}')
        except OSError:
            return None

        df = handle.read().to_pandas()
        numpy_data = self.prepare_data(data_frame=df)
        return numpy_data

    @staticmethod
    def prepare_data(data_frame: pd.DataFrame) -> np.array:
        """
        Reads footprint data from a parquet file.

        Returns: (np.array) footprint data loaded from the parquet file
        """
        areaperil_id = data_frame["areaperil_id"].to_numpy()
        intensity_bin_id = data_frame["intensity_bin_id"].to_numpy()
        probability = data_frame["probability"].to_numpy()

        buffer = np.empty(len(areaperil_id), dtype=Event)
        outcome = stitch_data(areaperil_id, intensity_bin_id, probability, buffer)
        return np.array(outcome, dtype=Event)


@nb.jit(cache=True)
def stitch_data(areaperil_id, intensity_bin_id, probability, buffer):
    """
    Creates a list of tuples from three np.arrays (all inputs must be the same length).
    Args:
        areaperil_id: (np.array) list of areaperil IDs
        intensity_bin_id: (np.array) list of probability bin IDs
        probability: (np.array) list of probabilities
        buffer: (np.array[Event]) list of zeros to be populated with the previous lists
    Returns:
    """
    for x in range(0, len(buffer)):
        buffer[x]['areaperil_id'] = areaperil_id[x]
        buffer[x]['intensity_bin_id'] = intensity_bin_id[x]
        buffer[x]['probability'] = probability[x]
    return buffer



import os
import numpy as np
import numba as nb

allowed_allocation_rule = [0, 1, 2, 3]

np_oasis_int = np.int32
nb_oasis_int = nb.int32

areaperil_int = os.environ.get('AREAPERIL_TYPE', 'u4')

np_oasis_float = np.float32
nb_oasis_float = nb.float32

float_equal_precision = np.finfo(np_oasis_float).eps


@nb.njit(cache=True)
def almost_equal(a, b):
    return abs(a - b) < float_equal_precision


null_index = np_oasis_int(-1)
need_tiv_policy = (4, 6, 18, 27, 28, 29, 30, 31, 32, 37)
need_extras = (7, 8, 10, 11, 13, 19, 26, 27, 35, 36)

# financial structure static input dtypes
fm_programme_dtype = np.dtype([('from_agg_id', 'i4'), ('level_id', 'i4'), ('to_agg_id', 'i4')])
fm_policytc_dtype = np.dtype([('level_id', 'i4'), ('agg_id', 'i4'), ('layer_id', 'i4'), ('policytc_id', 'i4')])
fm_profile_dtype = np.dtype([('policytc_id', 'i4'),
                             ('calcrule_id', 'i4'),
                             ('deductible_1', 'f4'),
                             ('deductible_2', 'f4'),
                             ('deductible_3', 'f4'),
                             ('attachment_1', 'f4'),
                             ('limit_1', 'f4'),
                             ('share_1', 'f4'),
                             ('share_2', 'f4'),
                             ('share_3', 'f4'),
                             ])
fm_profile_step_dtype = np.dtype([('policytc_id', 'i4'),
                                  ('calcrule_id', 'i4'),
                                  ('deductible_1', 'f4'),
                                  ('deductible_2', 'f4'),
                                  ('deductible_3', 'f4'),
                                  ('attachment_1', 'f4'),
                                  ('limit_1', 'f4'),
                                  ('share_1', 'f4'),
                                  ('share_2', 'f4'),
                                  ('share_3', 'f4'),
                                  ('step_id', 'i4'),
                                  ('trigger_start', 'f4'),
                                  ('trigger_end', 'f4'),
                                  ('payout_start', 'f4'),
                                  ('payout_end', 'f4'),
                                  ('limit_2', 'f4'),
                                  ('scale_1', 'f4'),
                                  ('scale_2', 'f4'),
                                  ])
fm_profile_csv_col_map = {
                         'deductible_1': 'deductible1',
                         'deductible_2': 'deductible2',
                         'deductible_3': 'deductible3',
                         'attachment_1': 'attachment1',
                         'limit_1': 'limit1',
                         'share_1': 'share1',
                         'share_2': 'share2',
                         'share_3': 'share3',
                         'limit_2':' limit2',
                         'scale_1':'scale1',
                         'scale_2': 'scale2',
                        }
fm_xref_dtype = np.dtype([('output_id', 'i4'), ('agg_id', 'i4'), ('layer_id', 'i4')])
fm_xref_csv_col_map = {'output_id':'output'}

coverages_dtype = np.dtype([('coverage_id', 'i4'), ('tiv', 'f4')])

items_dtype = np.dtype([('item_id', 'i4'),
                        ('coverage_id', 'i4'),
                        ('areaperil_id', areaperil_int),
                        ('vulnerability_id', 'i4'),
                        ('group_id', 'i4')])

EXTRA_VALUES = 3


__all__ = [
    'create_target_directory',
    'prepare_input_files_directory'
]

import filecmp
import os
import shutil

from pathlib import Path

from ..utils.exceptions import OasisException
from ..utils.path import as_path
from ..utils.defaults import store_exposure_fp


def create_target_directory(target_dir, label):
    target_dir = as_path(target_dir, label, is_dir=True, preexists=False)
    if not os.path.exists(target_dir):
        Path(target_dir).mkdir(parents=True, exist_ok=True)

    return target_dir


def prepare_input_files_directory(
    target_dir,
    exposure_fp,
    exposure_profile_fp=None,
    keys_fp=None,
    keys_errors_fp=None,
    lookup_config_fp=None,
    model_version_fp=None,
    complex_lookup_config_fp=None,
    accounts_fp=None,
    accounts_profile_fp=None,
    fm_aggregation_profile_fp=None,
    ri_info_fp=None,
    ri_scope_fp=None
):
    try:
        # Prepare the target directory and copy the source files, profiles and
        # model version file into it
        target_dir = create_target_directory(
            target_dir, 'target Oasis files directory'
        )

        # Copy preserving original filenames 
        paths = [
            (p, os.path.join(target_dir, os.path.basename(p))) for p in (
                exposure_profile_fp, accounts_profile_fp,
                fm_aggregation_profile_fp, lookup_config_fp, model_version_fp,
                keys_fp, keys_errors_fp
            ) if p
        ]

        # Copy and rename to default set in 
        # oasislmf.utils.defaults.SOURCE_FILENAMES
        paths_rename = (
            (exposure_fp, "loc"), 
            (accounts_fp, "acc"), 
            (ri_info_fp, "info"), 
            (ri_scope_fp, "scope"),
            (complex_lookup_config_fp, "complex_lookup")
        )    

        for fp, key in paths_rename:
            if fp:
                # check if exposure pre-analysis has run:
                if not os.path.exists(os.path.join(target_dir, f'epa_{store_exposure_fp(fp, key)}')):
                    paths.append((fp, os.path.join(target_dir, store_exposure_fp(fp, key))))

        for src, dst in paths:
            if src and os.path.exists(src):
                shutil.copy2(src, dst) if not (os.path.exists(dst) and filecmp.cmp(src, dst, shallow=False)) else None
    except (FileNotFoundError, IOError, OSError, shutil.Error, TypeError, ValueError) as e:
        raise OasisException("Exception raised in 'prepare_input_files_directory'", e)

    return target_dir



__all__ = [
    'KeyServerFactory',
    'BasicKeyServer'
]

import copy
import csv
import json
import os
import sys
import types
import warnings

from collections import OrderedDict
from contextlib import ExitStack

import math
import numpy as np
import pandas as pd

from ..utils.data import get_json, get_location_df
from ..utils.exceptions import OasisException
from ..utils.log import oasis_log
from ..utils.path import import_from_string, get_custom_module, as_path
from ..utils.status import OASIS_KEYS_STATUS

from .builtin import DeterministicLookup
from .builtin import Lookup as NewLookup

from multiprocessing import cpu_count,  Queue, Process
from queue import Empty, Full

# add pickling support for traceback object
import tblib.pickling_support
tblib.pickling_support.install()


def with_error_queue(fct):
    def wrapped_fct(error_queue, *args, **kwargs):
        try:
            return fct(error_queue, *args, **kwargs)
        except Exception:
            error_queue.put(sys.exc_info())
    return wrapped_fct


class KeyServerFactory(object):
    """
    A factory class to create the Keys Server that will be use to generate the keys files
    All Key Server must implement the interface defined in lookup.interface.KeyServerInterface

    Oasis provides a built-in Key Server that manage the generation of the key files from the key provided by
    a built-in or a custom Key Lookup.

    The factory now return a KeyServer object and not a KeyLookup.
    The parameter to pass has also been simplified
    usage of all the below parameter are now deprecated
      - complex_lookup_config_fp => pass the path to your complex lookup config directly in lookup_config_fg
      - lookup_module_path => set as key 'lookup_module_path' in the lookup config
      - model_keys_data_path => set as key 'keys_data_path' in the lookup config
      - model_version_file_path => set the model information ('supplier_id', 'model_id', 'model_version') directly
        into the config
    """

    @classmethod
    def get_config(cls, config_fp):
        return as_path(os.path.dirname(config_fp), 'config_fp'), get_json(config_fp)

    @classmethod
    def get_model_info(cls, model_version_file_path):
        """
        Get model information from the model version file.
        """
        model_version_file_path = as_path(model_version_file_path, 'model_version_file_path', preexists=True, null_is_valid=False)

        with open(model_version_file_path, 'r', encoding='utf-8') as f:
            return next(csv.DictReader(
                f, fieldnames=['supplier_id', 'model_id', 'model_version']
            ))

    @classmethod
    def update_deprecated_args(cls, config_dir, config,
                               complex_lookup_config_fp, model_keys_data_path, model_version_file_path, lookup_module_path):

        if (complex_lookup_config_fp
                or model_keys_data_path
                or model_version_file_path
                or lookup_module_path):
            warnings.warn('usage of complex_lookup_config_fp, model_keys_data_path, '
                          'model_version_file_path and lookup_module_path is now deprecated'
                          'those variables now need to be set in lookup config see (key server documentation)')

        if complex_lookup_config_fp:
            config_dir, config = cls.get_config(complex_lookup_config_fp)

        if model_keys_data_path:
            config['keys_data_path'] = as_path(model_keys_data_path, 'model_keys_data_path', preexists=True)

        if model_version_file_path:
            config['model'] = cls.get_model_info(model_version_file_path)

        if lookup_module_path:
            config['lookup_module_path'] = lookup_module_path

        return config_dir, config

    @classmethod
    def create(
        cls,
        model_keys_data_path=None,
        model_version_file_path=None,
        lookup_module_path=None,
        lookup_config=None,
        lookup_config_json=None,
        lookup_config_fp=None,
        complex_lookup_config_fp=None,
        user_data_dir=None,
        output_directory=None,
    ):
        """
        Creates a keys lookup class instance for the given model and supplier -
        local file paths are required for the model keys data folder, the model
        version file and the Git repository for the model keys server. Returns a
        pair ``(model_info, klc)``, where ``model_info`` is a dictionary holding
        model information from the model version file and `klc` is the lookup
        service class instance for the model.
        """
        if lookup_config:
            config_dir = '.'
            config = lookup_config
        elif lookup_config_json:
            config_dir = '.'
            config = json.loads(lookup_config_json)
        elif lookup_config_fp:
            config_dir, config = cls.get_config(lookup_config_fp)
        else: # no config
            config_dir, config = '.', {}

        if not config:
            config_dir, config = cls.update_deprecated_args(config_dir, config,
                                                            complex_lookup_config_fp, model_keys_data_path,
                                                            model_version_file_path, lookup_module_path)
        else: # reproduce lookup_config overwrite complex_lookup_config_fp
            complex_lookup_config_fp = None

        if config.get('key_server_module_path'):
            _KeyServer_module = get_custom_module(config.get('key_server_module_path'), 'key_server_module_path')
            _KeyServer = getattr(_KeyServer_module, '{}KeysServer'.format(config['model']['model_id']))
        else:
            _KeyServer = BasicKeyServer

        if _KeyServer.interface_version == '1':
            key_server = _KeyServer(config,
                                    config_dir=config_dir,
                                    user_data_dir=user_data_dir,
                                    output_dir=output_directory)
        else:
            raise OasisException(f"KeyServer interface version {_KeyServer.interface_version} not implemented")

        if complex_lookup_config_fp:
            key_server.complex_lookup_config_fp = complex_lookup_config_fp

        return config['model'], key_server


class BasicKeyServer:
    """
    A basic implementation of the KeyServerInterface
    will load the KeyLookup class from config['lookup_module_path'] if present or used the built-in KeyLookup
    KeyLookup must implement the KeyLookupInterface

    will provide a multiprocess solution if KeyLoopup implement the process_locations_multiproc method

    both single and multiprocess solutions will use low amount of memory
    as they process the key by chunk of limited size.

    This class implement all the file writing method that were previously handled by the lookup factory
    """
    interface_version = "1"

    valid_format = ['oasis', 'json']

    error_heading_row = OrderedDict([
        ('loc_id', 'LocID'),
        ('peril_id', 'PerilID'),
        ('coverage_type', 'CoverageTypeID'),
        ('status', 'Status'),
        ('message', 'Message'),
    ])

    model_data_heading_row = OrderedDict([
        ('loc_id', 'LocID'),
        ('peril_id', 'PerilID'),
        ('coverage_type', 'CoverageTypeID'),
        ('model_data', 'ModelData'),
    ])

    key_success_heading_row = OrderedDict([
        ('loc_id', 'LocID'),
        ('peril_id', 'PerilID'),
        ('coverage_type', 'CoverageTypeID'),
        ('area_peril_id', 'AreaPerilID'),
        ('vulnerability_id', 'VulnerabilityID'),
    ])

    key_success_with_message_heading_row = OrderedDict([
        ('loc_id', 'LocID'),
        ('peril_id', 'PerilID'),
        ('coverage_type', 'CoverageTypeID'),
        ('area_peril_id', 'AreaPerilID'),
        ('vulnerability_id', 'VulnerabilityID'),
        ('message', 'Message')
    ])

    min_bloc_size = 1000
    max_bloc_size = 10000

    def __init__(self, config, config_dir=None, user_data_dir=None, output_dir=None):
        self.config = config
        self.config_dir = config_dir or '.'
        self.user_data_dir = user_data_dir
        self.output_dir = output_dir
        self.lookup_cls = self.get_lookup_cls()

    def get_lookup_cls(self):
        if self.config.get('lookup_class'):
            lookup_cls = import_from_string(self.config.get('lookup_class'))

        elif self.config.get('lookup_module'):
            lookup_module = import_from_string(self.config.get('lookup_module'))
            lookup_cls = getattr(lookup_module, '{}KeysLookup'.format(self.config['model']['model_id']))

        elif self.config.get('lookup_module_path'):
            lookup_module_path = self.config.get('lookup_module_path')
            if not os.path.isabs(lookup_module_path):
                lookup_module_path = os.path.join(self.config_dir, lookup_module_path)
            lookup_module = get_custom_module(lookup_module_path, 'lookup_module_path')
            lookup_cls = getattr(lookup_module, '{}KeysLookup'.format(self.config['model']['model_id']))
        else: # built-in lookup
            if self.config.get('builtin_lookup_type') == 'deterministic':
                lookup_cls = DeterministicLookup
            elif self.config.get('builtin_lookup_type') == 'new_lookup':
                lookup_cls = NewLookup
            else:
                raise OasisException(f"Unrecognised lookup config file, or config file is from deprecated built in lookup module 'oasislmf<=1.16.0' ")

        return lookup_cls

    @staticmethod
    def create_lookup(lookup_cls, config, config_dir, user_data_dir, output_dir, lookup_id):
        lookup_config = copy.deepcopy(config)
        lookup_config['lookup_id'] = lookup_id
        lookup_interface_version = getattr(lookup_cls, 'interface_version', '0')
        if lookup_interface_version == '1':
            return lookup_cls(config,
                              config_dir=config_dir,
                              user_data_dir=user_data_dir,
                              output_dir=output_dir)
        elif lookup_interface_version == '0':
            warnings.warn('OasisLookupInterface (or OasisBaseKeysLookup) is now deprecated'
                          ' Interface for lookup is now lookup.interface.LookupInterface'
                          ' for similar functionality use lookup.base.AbstractBasicKeyLookup'
                          ' for multiprocess implementation add lookup.base.MultiprocLookupMixin')
            if not (config and output_dir):
                return lookup_cls(
                    keys_data_directory=config.get('keys_data_path'),
                    supplier=config['model']['supplier_id'],
                    model_name=config['model']['model_id'],
                    model_version=config['model']['model_version'],
                )
            elif not user_data_dir:
                return lookup_cls(
                    keys_data_directory=config.get('keys_data_path'),
                    supplier=config['model']['supplier_id'],
                    model_name=config['model']['model_id'],
                    model_version=config['model']['model_version'],
                    complex_lookup_config_fp=config_dir,
                    output_directory=output_dir
                )
            else:
                return lookup_cls(
                    keys_data_directory=config.get('keys_data_path'),
                    supplier=config['model']['supplier_id'],
                    model_name=config['model']['model_id'],
                    model_version=config['model']['model_version'],
                    complex_lookup_config_fp=config_dir,
                    user_data_dir=user_data_dir,
                    output_directory=output_dir
                )
        else:
            raise OasisException(f"lookup interface version {lookup_interface_version} not implemented")

    def get_locations(self, location_fp):
        """load exposure data from location_fp and return the exposure dataframe"""
        return get_location_df(location_fp)

    @staticmethod
    @with_error_queue
    def location_producer(error_queue, loc_df, part_count, loc_queue):
        loc_ids_parts = np.array_split(np.unique(loc_df['loc_id']), part_count)
        loc_df_parts = (loc_df[loc_df['loc_id'].isin(loc_ids_parts[i])] for i in range(part_count))
        loc_df_part = True
        while loc_df_part is not None:
            loc_df_part = next(loc_df_parts, None)
            while error_queue.empty():
                try:
                    loc_queue.put(loc_df_part, timeout=5)
                    break
                except Full:
                    pass
            else:
                return

    @staticmethod
    @with_error_queue
    def lookup_multiproc_worker(error_queue, lookup_cls, config, config_dir, user_data_dir, output_dir, lookup_id, loc_queue, key_queue):
        lookup = BasicKeyServer.create_lookup(lookup_cls, config, config_dir, user_data_dir, output_dir, lookup_id)
        while True:
            while error_queue.empty():
                try:
                    loc_df_part = loc_queue.get(timeout=5)

                    break
                except Empty:
                    pass
            else:
                return

            if loc_df_part is None:
                loc_queue.put(None)
                key_queue.put(None)
                break

            while error_queue.empty():
                try:
                    key_queue.put(lookup.process_locations_multiproc(loc_df_part), timeout=5)
                    break
                except Full:
                    pass
            else:
                return

    @staticmethod
    def key_producer(key_queue, error_queue, worker_count):
        finished_workers = 0
        while finished_workers < worker_count and error_queue.empty():
            while error_queue.empty():
                try:
                    res = key_queue.get(timeout=5)
                    break
                except Empty:
                    pass
            else:
                break

            if res is None:
                finished_workers+=1
            else:
                yield res

    def get_success_heading_row(self, keys, keys_success_msg):
        if 'model_data' in keys:
            return self.model_data_heading_row
        elif keys_success_msg:
            return self.key_success_with_message_heading_row
        else:
            return self.key_success_heading_row

    def write_json_keys_file(self, results, keys_success_msg, successes_fp, errors_fp):
        # no streaming implementation for json format
        results = pd.concat((r for r in results if not r.empty))

        success = results['status'] == OASIS_KEYS_STATUS['success']['id']
        success_df = results[success]
        success_df.to_json(successes_fp, orient='records', indent=4, force_ascii=False)
        successes_count = success_df.shape[0]
        if errors_fp:
            errors_df = results[~success]
            errors_df.to_json(errors_fp, orient='records', indent=4, force_ascii=False)
            error_count = errors_df.shape[0]
        else:
            error_count = 0
        return successes_count, error_count

    def write_oasis_keys_file(self, results, keys_success_msg, successes_fp, errors_fp):
        with ExitStack() as stack:
            successes_file = stack.enter_context(open(successes_fp, 'w', encoding='utf-8'))
            if errors_fp:
                errors_file = stack.enter_context(open(errors_fp, 'w', encoding='utf-8'))
                errors_file.write(','.join(self.error_heading_row.values()) + '\n')
            else:
                errors_file = None
            success_heading_row = None
            successes_count = 0
            error_count = 0
            for i, result in enumerate(results):
                success = result['status'] == OASIS_KEYS_STATUS['success']['id']
                success_df = result[success]
                if success_heading_row is None:
                    success_heading_row = self.get_success_heading_row(result.columns, keys_success_msg)
                success_df[success_heading_row.keys()].rename(columns=success_heading_row
                                                                ).to_csv(successes_file, index=False, header=not i)
                successes_count += success_df.shape[0]
                if errors_file:
                    errors_df = result[~success]
                    if 'message' not in errors_df.columns:
                        errors_df['message'] = "" # If no error message column, fill with blank to prevent KeyError
                    errors_df[self.error_heading_row.keys()].rename(columns=self.error_heading_row
                                                                    ).to_csv(errors_file, index=False, header=False)
                    error_count += errors_df.shape[0]
            return successes_count, error_count

    def write_keys_file(self, results, successes_fp, errors_fp, output_format, keys_success_msg):
        if output_format not in self.valid_format:
            raise OasisException(f"Unrecognised lookup file output format {output_format} - valid formats are {self.valid_format}")

        write = getattr(self, f'write_{output_format}_keys_file')
        successes_count, error_count = write(results, keys_success_msg, successes_fp, errors_fp)

        if errors_fp:
            return successes_fp, successes_count, errors_fp, error_count
        else:
            return successes_fp, successes_count

    def generate_key_files_singleproc(self, loc_df, successes_fp, errors_fp, output_format, keys_success_msg, **kwargs):
        if getattr(self, 'complex_lookup_config_fp', None):  # backward compatibility 1.15 hack
            config_dir = getattr(self, 'complex_lookup_config_fp', None)
        else:
            config_dir = self.config_dir
        lookup = self.create_lookup(self.lookup_cls, self.config, config_dir, self.user_data_dir, self.output_dir,
                                    lookup_id=None)

        key_results = lookup.process_locations(loc_df)

        def gen_results(results):
            if isinstance(results, pd.DataFrame):
                yield results
            elif isinstance(results, (list, tuple)):
                yield pd.DataFrame(results)
            elif isinstance(results, types.GeneratorType):
                results_part = pd.DataFrame.from_records(results, nrows=self.max_bloc_size)
                while not results_part.empty:
                    yield results_part
                    results_part = pd.DataFrame.from_records(results, nrows=self.max_bloc_size)
            else:
                raise OasisException("Unrecognised type for results: {type(results)}. expected ")

        return self.write_keys_file(gen_results(key_results),
                                    successes_fp=successes_fp,
                                    errors_fp=errors_fp,
                                    output_format=output_format,
                                    keys_success_msg=keys_success_msg,)

    def generate_key_files_multiproc(self, loc_df, successes_fp, errors_fp, output_format, keys_success_msg,
                                     num_cores, num_partitions, **kwargs):
        """
        Process and return the lookup results a location row
        Used in multiprocessing based query

        location_row is of type <class 'pandas.core.series.Series'>

        """
        if getattr(self, 'complex_lookup_config_fp', None):  # backward compatibility 1.15 hack
            config_dir = getattr(self, 'complex_lookup_config_fp', None)
        else:
            config_dir = self.config_dir

        pool_count = num_cores if num_cores > 0 else cpu_count()
        if num_partitions > 0:
            part_count = num_partitions
        else:
            bloc_size = min(max(math.ceil(loc_df.shape[0] / pool_count), self.min_bloc_size), self.max_bloc_size)
            part_count = math.ceil(loc_df.shape[0] / bloc_size)
            pool_count = min(pool_count, part_count)
        if pool_count <= 1:
            return self.generate_key_files_singleproc(loc_df, successes_fp, errors_fp, output_format, keys_success_msg)

        loc_queue = Queue(maxsize=pool_count)
        key_queue = Queue(maxsize=pool_count)
        error_queue = Queue()

        location_producer = Process(target=self.location_producer, args=(error_queue, loc_df, part_count, loc_queue))

        workers = [Process(target=self.lookup_multiproc_worker,
                           args=(error_queue, self.lookup_cls, self.config, config_dir,
                                 self.user_data_dir, self.output_dir,
                                 lookup_id, loc_queue, key_queue))
                   for lookup_id in range(pool_count)]

        location_producer.start()
        [worker.start() for worker in workers]

        try:
            return self.write_keys_file(self.key_producer(key_queue, error_queue, worker_count= pool_count),
                                        successes_fp=successes_fp,
                                        errors_fp=errors_fp,
                                        output_format=output_format,
                                        keys_success_msg=keys_success_msg,)
        except Exception:
            error_queue.put(sys.exc_info())
        finally:
            for process in [location_producer] + workers:
                if process.is_alive():
                    process.terminate()
                    process.join()
            loc_queue.close()
            key_queue.close()
            if not error_queue.empty():
                exc_info = error_queue.get()
                raise exc_info[0].with_traceback(exc_info[1], exc_info[2])

    @oasis_log()
    def generate_key_files(
        self,
        location_fp,
        successes_fp,
        errors_fp=None,
        output_format='oasis',
        keys_success_msg=False,
        multiproc_enabled=True,
        multiproc_num_cores=-1,
        multiproc_num_partitions=-1,
        location_df=None,
        **kwargs
    ):
        """
        generate key files by calling:
        1. get_locations to get a location object from the location_fp
        2. process_locations or process_locations_multiproc to get results object from the locations object
        3. write_keys_file to writes the relevant files from the results object
        """
        successes_fp = as_path(successes_fp, 'successes_fp', preexists=False)
        errors_fp = as_path(errors_fp, 'errors_fp', preexists=False)

        if location_df is not None:
            locations = location_df
        else:
            locations = self.get_locations(location_fp)

        if multiproc_enabled and hasattr(self.lookup_cls, 'process_locations_multiproc'):
            return self.generate_key_files_multiproc(locations,
                                                     successes_fp=successes_fp,
                                                     errors_fp=errors_fp,
                                                     output_format=output_format,
                                                     keys_success_msg=keys_success_msg,
                                                     num_cores = multiproc_num_cores,
                                                     num_partitions = multiproc_num_partitions)
        else:
            return self.generate_key_files_singleproc(locations,
                                                      successes_fp=successes_fp,
                                                      errors_fp=errors_fp,
                                                      output_format=output_format,
                                                      keys_success_msg=keys_success_msg,
                                                      )
